@@ -1,8 +1,9 @@
+#include <string.h>
+
 #include <R.h>
 #include <Rdefines.h>
 #include <Rinternals.h>
-
-#include <string.h>
+#include <R_ext/Riconv.h>
 
 #include "compile.h"
 #include "match.h"
@@ -13,6 +14,9 @@ extern UChar * onigenc_step (OnigEncoding enc, const UChar *p, const UChar *end,
 // Block size for match data; memory is allocated in chunks this big
 #define MATCH_BLOCK_SIZE    128
 
+// Initial buffer size when reading from a file; scales exponentially
+#define FILE_BUFFER_SIZE    1024
+
 // Allocate memory for a rawmatch_t object with capacity one block, and its contents
 rawmatch_t * ore_rawmatch_alloc (const int n_regions)
 {
@@ -74,7 +78,7 @@ void ore_rawmatch_store_string (rawmatch_t *match, const size_t loc, const char
 }
 
 // Search a single string for matches to a regex
-rawmatch_t * ore_search (regex_t *regex, const char *text, const Rboolean all, const size_t start)
+rawmatch_t * ore_search (regex_t *regex, const char *text, const char *text_end, const Rboolean all, const size_t start)
 {
     int return_value, length;
     rawmatch_t *result = NULL;
@@ -85,11 +89,19 @@ rawmatch_t * ore_search (regex_t *regex, const char *text, const Rboolean all, c
     // The number of matches found so far
     int match_number = 0;
     
+    // Find the end-point of the text (for binary data it may include null bytes)
+    UChar *end_ptr;
+    if (text_end != NULL)
+        end_ptr = (UChar *) text_end;
+    else
+        end_ptr = (UChar *) text + strlen(text);
+    
     // If we're not starting at the beginning, step forward the required number of characters
-    UChar *end_ptr = (UChar *) text + strlen(text);
     UChar *start_ptr;
     if (start == 0)
         start_ptr = (UChar *) text;
+    else if (regex->enc->max_enc_len == 1)
+        start_ptr = (UChar *) text + start;
     else
         start_ptr = onigenc_step(regex->enc, (UChar *) text, end_ptr, (int) start);
     
@@ -182,16 +194,46 @@ void ore_int_vector (SEXP vec, const int *data, const int n_regions, const int n
         ptr[i] = data[i*n_regions] + increment;
 }
 
-// Copy string data from a rawmatch_t to an R vector
-void ore_char_vector (SEXP vec, const char **data, const int n_regions, const int n_matches, const cetype_t encoding)
+// Wrapper around Riconv, to convert between encodings
+const char * ore_iconv (void *iconv_handle, const char *old)
 {
+    if (iconv_handle)
+    {
+        size_t old_size = strlen(old);
+        size_t new_size = old_size * 6;
+        char *buffer = R_alloc(new_size+1, 1);
+        char *buffer_start = buffer;
+        Riconv(iconv_handle, &old, &old_size, &buffer, &new_size);
+        *buffer = '\0';
+        return buffer_start;
+    }
+    else
+        return old;
+}
+
+// Copy string data from a rawmatch_t to an R vector
+void ore_char_vector (SEXP vec, const char **data, const int n_regions, const int n_matches, cetype_t encoding, const char *old_enc_name)
+{
+    void *iconv_handle = NULL;
+    if (old_enc_name != NULL)
+    {
+        if (ore_strnicmp(old_enc_name, "native.enc", 10) == 0)
+            iconv_handle = Riconv_open("UTF-8", "");
+        else
+            iconv_handle = Riconv_open("UTF-8", old_enc_name);
+        encoding = CE_UTF8;
+    }
+    
     for (int i=0; i<n_matches; i++)
     {
         if (data[i*n_regions] == NULL)
             SET_STRING_ELT(vec, i, mkCharCE("",encoding));
         else
-            SET_STRING_ELT(vec, i, mkCharCE(data[i*n_regions],encoding));
+            SET_STRING_ELT(vec, i, mkCharCE(ore_iconv(iconv_handle,data[i*n_regions]), encoding));
     }
+    
+    if (iconv_handle)
+        Riconv_close(iconv_handle);
 }
 
 // Copy integer data for groups into an R matrix
@@ -218,8 +260,18 @@ void ore_int_matrix (SEXP mat, const int *data, const int n_regions, const int n
 }
 
 // Copy string data from groups into an R matrix
-void ore_char_matrix (SEXP mat, const char **data, const int n_regions, const int n_matches, const SEXP col_names, const cetype_t encoding)
+void ore_char_matrix (SEXP mat, const char **data, const int n_regions, const int n_matches, const SEXP col_names, cetype_t encoding, const char *old_enc_name)
 {
+    void *iconv_handle = NULL;
+    if (old_enc_name != NULL)
+    {
+        if (ore_strnicmp(old_enc_name, "native.enc", 10) == 0)
+            iconv_handle = Riconv_open("UTF-8", "");
+        else
+            iconv_handle = Riconv_open("UTF-8", old_enc_name);
+        encoding = CE_UTF8;
+    }
+    
     for (int i=0; i<n_matches; i++)
     {
         for (int j=1; j<n_regions; j++)
@@ -229,10 +281,13 @@ void ore_char_matrix (SEXP mat, const char **data, const int n_regions, const in
             if (element == NULL)
                 SET_STRING_ELT(mat, (j-1)*n_matches + i, NA_STRING);
             else
-                SET_STRING_ELT(mat, (j-1)*n_matches + i, mkCharCE(element,encoding));
+                SET_STRING_ELT(mat, (j-1)*n_matches + i, mkCharCE(ore_iconv(iconv_handle,element), encoding));
         }
     }
     
+    if (iconv_handle)
+        Riconv_close(iconv_handle);
+    
     // Set column names if supplied
     if (!isNull(col_names))
     {
@@ -246,18 +301,66 @@ void ore_char_matrix (SEXP mat, const char **data, const int n_regions, const in
     }
 }
 
+// Read text from a file into a buffer
+file_contents_t * ore_read_file (const char *filename)
+{
+    FILE *fp = fopen(filename, "rb");
+    if (fp == NULL)
+        error("Could not open file %s", filename);
+    
+    size_t old_buffer_size = 0;
+    size_t buffer_size = FILE_BUFFER_SIZE;
+    char *buffer = (char *) R_alloc(buffer_size, 1);
+    char *ptr = buffer;
+    
+    while (TRUE)
+    {
+        size_t n = buffer_size - old_buffer_size;
+        size_t bytes_read = fread(ptr, 1, n, fp);
+        if (bytes_read < n)
+        {
+            ptr += bytes_read;
+            break;
+        }
+        else
+        {
+            old_buffer_size = buffer_size;
+            buffer_size *= 2;
+            buffer = (char *) ore_realloc(buffer, buffer_size, old_buffer_size, 1);
+            ptr = buffer + old_buffer_size;
+        }
+    }
+    
+    fclose(fp);
+    
+    file_contents_t *contents = (file_contents_t *) R_alloc(1, sizeof(file_contents_t));
+    contents->start = buffer;
+    contents->end = ptr;
+    return contents;
+}
+
 // Vectorised wrapper around ore_search(), which handles the R API stuff
 SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simplify_)
 {
     // Convert R objects to C types
-    regex_t *regex = ore_retrieve(regex_, text_);
     SEXP group_names = getAttrib(regex_, install("groupNames"));
     const Rboolean all = asLogical(all_) == TRUE;
     const Rboolean simplify = asLogical(simplify_) == TRUE;
     int *start = INTEGER(start_);
     
+    // Check whether the text argument is actually a file path
+    Rboolean using_file = FALSE;
+    SEXP textClass = getAttrib(text_, R_ClassSymbol);
+    if (length(textClass) > 0 && strcmp(CHAR(STRING_ELT(textClass,0)), "orefile") == 0)
+        using_file = TRUE;
+    else
+        PROTECT(text_ = AS_CHARACTER(text_));
+    
+    // Retrieve the regex
+    regex_t *regex = ore_retrieve(regex_, text_, using_file);
+    
     // Obtain the lengths of the text and start vectors (the latter will be recycled if necessary)
-    const int text_len = length(text_);
+    const int text_len = using_file ? 1 : length(text_);
     const int start_len = length(start_);
     
     // Check for sensible input
@@ -272,16 +375,38 @@ SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simpl
     // Step through each string to be searched
     for (int i=0; i<text_len; i++)
     {
-        const cetype_t encoding = getCharCE(STRING_ELT(text_, i));
-        if ((encoding == CE_UTF8 && regex->enc == ONIG_ENCODING_ISO_8859_1) || (encoding == CE_LATIN1 && regex->enc == ONIG_ENCODING_UTF8))
-        {
-            warning("Encoding of text element %d does not match the regex", i+1);
-            SET_ELEMENT(results, i, R_NilValue);
-            continue;
-        }
+        rawmatch_t *raw_match;
+        cetype_t encoding = CE_NATIVE;
+        file_contents_t *contents;
+        const char *file_encoding_string;
         
-        // Do the match
-        rawmatch_t *raw_match = ore_search(regex, CHAR(STRING_ELT(text_,i)), all, (size_t) start[i % start_len] - 1);
+        if (using_file)
+        {
+            // Check the file's encoding
+            SEXP encoding_name = getAttrib(text_, install("encoding"));
+            file_encoding_string = CHAR(STRING_ELT(encoding_name, 0));
+            OnigEncoding file_encoding = ore_name_to_onig_enc(file_encoding_string);
+            if (file_encoding != regex->enc)
+                warning("File encoding does not match the regex");
+            
+            // Do the match
+            contents = ore_read_file(CHAR(STRING_ELT(text_, 0)));
+            raw_match = ore_search(regex, contents->start, contents->end, all, (size_t) start[0] - 1);
+        }
+        else
+        {
+            // Check the string's encoding
+            encoding = getCharCE(STRING_ELT(text_, i));
+            if ((encoding == CE_UTF8 && regex->enc == ONIG_ENCODING_ISO_8859_1) || (encoding == CE_LATIN1 && regex->enc == ONIG_ENCODING_UTF8))
+            {
+                warning("Encoding of text element %d does not match the regex", i+1);
+                SET_ELEMENT(results, i, R_NilValue);
+                continue;
+            }
+        
+            // Do the match
+            raw_match = ore_search(regex, CHAR(STRING_ELT(text_,i)), NULL, all, (size_t) start[i % start_len] - 1);
+        }
         
         // Assign NULL if there's no match, otherwise build up an "orematch" object
         if (raw_match == NULL)
@@ -304,7 +429,11 @@ SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simpl
             SET_STRING_ELT(result_names, 6, mkChar("matches"));
             
             // Convert elements of the raw match data to R vectors
-            PROTECT(text = ScalarString(STRING_ELT(text_,i)));
+            // Note that the text can't easily be returned from a file because offsets and lengths may be wrong after translation between encodings
+            if (using_file)
+                text = R_NilValue;
+            else
+                PROTECT(text = ScalarString(STRING_ELT(text_,i)));
             PROTECT(n_matches = ScalarInteger(raw_match->n_matches));
             PROTECT(offsets = NEW_INTEGER(raw_match->n_matches));
             ore_int_vector(offsets, raw_match->offsets, raw_match->n_regions, raw_match->n_matches, 1);
@@ -315,7 +444,10 @@ SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simpl
             PROTECT(byte_lengths = NEW_INTEGER(raw_match->n_matches));
             ore_int_vector(byte_lengths, raw_match->byte_lengths, raw_match->n_regions, raw_match->n_matches, 0);
             PROTECT(matches = NEW_CHARACTER(raw_match->n_matches));
-            ore_char_vector(matches, (const char **) raw_match->matches, raw_match->n_regions, raw_match->n_matches, encoding);
+            if (using_file)
+                ore_char_vector(matches, (const char **) raw_match->matches, raw_match->n_regions, raw_match->n_matches, encoding, file_encoding_string);
+            else
+                ore_char_vector(matches, (const char **) raw_match->matches, raw_match->n_regions, raw_match->n_matches, encoding, NULL);
             
             // Put everything in place
             SET_ELEMENT(result, 0, text);
@@ -327,7 +459,7 @@ SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simpl
             SET_ELEMENT(result, 6, matches);
             
             // Unprotect everything back to "text"
-            UNPROTECT(7);
+            UNPROTECT(using_file ? 6 : 7);
             
             // If there are groups present, extract them
             if (raw_match->n_regions > 1)
@@ -353,7 +485,10 @@ SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simpl
                 PROTECT(byte_lengths = allocMatrix(INTSXP, raw_match->n_matches, raw_match->n_regions-1));
                 ore_int_matrix(byte_lengths, raw_match->byte_lengths, raw_match->n_regions, raw_match->n_matches, group_names, 0);
                 PROTECT(matches = allocMatrix(STRSXP, raw_match->n_matches, raw_match->n_regions-1));
-                ore_char_matrix(matches, (const char **) raw_match->matches, raw_match->n_regions, raw_match->n_matches, group_names, encoding);
+                if (using_file)
+                    ore_char_matrix(matches, (const char **) raw_match->matches, raw_match->n_regions, raw_match->n_matches, group_names, encoding, file_encoding_string);
+                else
+                    ore_char_matrix(matches, (const char **) raw_match->matches, raw_match->n_regions, raw_match->n_matches, group_names, encoding, NULL);
                 
                 // Put everything in place
                 SET_ELEMENT(groups, 0, offsets);
@@ -378,7 +513,7 @@ SEXP ore_search_all (SEXP regex_, SEXP text_, SEXP all_, SEXP start_, SEXP simpl
         }
     }
     
-    UNPROTECT(1);
+    UNPROTECT(using_file ? 1 : 2);
     
     // Return just the first (and only) element of the full list, if requested
     if (simplify && text_len == 1)
